/**************************************************************************
* iptrkg- log grapher (iptrk - ip tracker)
* Copyright (C) 2003 Jason Nunn <jsno@leal.com.au>
* Darwin, NT, Australia.
* Originally developed for Arafura Connect (Dec 2000 - Mar 2003)
* Donated to GPL before liquidation.
*
* =======================================================================
* produces address domain histograms of ip_log log files
*
**************************************************************************/
#include <math.h>
#include <gd.h>
#include "defs.h"
#include "lib.p"
#include "lib_cgi.p"
#include "iptrkg_defs.h"
#include "iptrkg.p"

#define BAR_LENGTH 400

typedef struct    /*addr domain*/
{
  U_32    addr;
  voldr_t voldr;
} ad_t;

typedef struct    /*time domain*/
{
  time_t  time;
  treap_t *ad_bt;
} td_t;

typedef struct
{
  char   fn[STR_LEN + 1];
  time_t t;
  double v;
} html_t;

PRIVATE char    ip_log_path[STR_LEN + 1] = IPTRK_LOG_IP_PATHC;
PRIVATE time_t  start_time,
                end_time;
PRIVATE treap_t *td_bt;
PRIVATE ll_t    *html_ll;

/**************************************************************************
*
**************************************************************************/
PRIVATE void ad_add_entry(treap_t **bt, U_32 addr, voldr_t *v)
{
  ad_t e;

  int append_cmp_func(void *n1, void *n2)
  {
    return ((ad_t *)n1)->addr - ((ad_t *)n2)->addr;
  }

  e.addr = addr;
  memcpy(&e.voldr, v, sizeof(voldr_t));
  treap_append(bt, &e, sizeof(ad_t), append_cmp_func);
}

/**************************************************************************
*
**************************************************************************/
PRIVATE void load_ip_data(char *pathname, char *addr_str)
{
  FILE           *fp;
  char           str[STR_LEN + 1], *s;
  td_t           e, *p;
  voldr_t        v;
  struct in_addr in;

  int append_cmp_func(void *n1, void *n2)
  {
    return cmp_32(((td_t *)n1)->time, ((td_t *)n2)->time);
  }

  int find_cmp_func(void *n)
  {
    return cmp_32(e.time, ((td_t *)n)->time);
  }

 /*
  * start
  */
  fp = fopen(pathname, "r");
  if(fp == NULL)
  {
    log_e(__func__, "::fopen(pathname)");
    exit(-1);
  }

  if(log_ip_fast_find(fp, start_time))
    for(;;)
    {
      if(feof(fp))
        break;

      if(fgets(str, STR_LEN, fp) == NULL)
        break;

      if(str[0] != 'o')
        continue;

      if(sscanf(str, "%*c %lx", &(e.time)) != 1)
        continue;

      if(e.time < start_time || e.time > end_time)
        continue;

      if(fgets(str, STR_LEN, fp) == NULL)
        break;

      s = str;
      if(!get_next_vol_u64(&s, &(v.I))) continue;
      if(!get_next_vol_u64(&s, &(v.O))) continue;
      if(!get_next_vol_u64(&s, &(v.u))) continue;
      if(!get_next_vol_u64(&s, &(v.i))) continue;
      if(!get_next_vol_u64(&s, &(v.o))) continue;

      p = treap_find(&td_bt, find_cmp_func);
      if(p == NULL)
      {
        e.ad_bt = NULL;
        treap_append(&td_bt, &e, sizeof(td_t), append_cmp_func);
        p = treap_find(&td_bt, find_cmp_func);
        if(p == NULL)
        {
          log_m('E', __func__, "::treap_find(): failed");
          exit(-1);
        }
      }

      inet_aton(addr_str, &in);
      ad_add_entry(&(p->ad_bt), ntohl(in.s_addr), &v);
    }

  fclose(fp);
}

/**************************************************************************
*
**************************************************************************/
PRIVATE void parse_logs(void)
{
  char          filepath[STR_LEN + 1];
  DIR           *dp;
  struct dirent *d_st;

  dp = opendir(ip_log_path);
  if(dp == NULL)
  {
    log_e(__func__, "::opendir()");
    return;
  }

  printf("Loading Logs.\n");

  for(;;)
  {
    d_st = readdir(dp);
    if(d_st == NULL)
      break;

    if(!strncmp(d_st->d_name, "log.", 4))
    {
      sprintf(filepath, "%s/%s", ip_log_path, d_st->d_name);
      load_ip_data(filepath, d_st->d_name + 4);
    }
  }

  closedir(dp);
}

/**************************************************************************
*
**************************************************************************/
PRIVATE void make_a_graph(treap_t *bt, time_t t, char *filename)
{
  U_64           x;
  int            a1, a2,
                 h_pos;
  FILE           *fp;
  char           str[STR_LEN + 1];
  struct in_addr in;
  html_t         html_entry;

  void find_max(treap_t *t)
  {
    ad_t *n;

    if(t == NULL)
      return;

    find_max(t->l);

    no_entries++;
    n = TREAP_ENTRY(t);
    x = n->voldr.I + n->voldr.O + n->voldr.u + n->voldr.i + n->voldr.o;
    if(x > max_vol)
      max_vol = x;

    find_max(t->r);
  }

  void draw_line(U_64 len, int col)
  {
    if(!len)
      return;
    a1 = a2;
    a2 += scale_v(max_vol, BAR_SIZE, len);
    gdImageLine(image, a1, h_pos, a2, h_pos, col);
  }

  inline void draw_leg(char *l, int col)
  {
    gdImageFilledRectangle(image, a1, 55, a1 + 10, 65, col);
    title_font(image, 14, a1 + 15, 65, col, l);
    a1 += 40;
  }

  void w_ad(treap_t *t)
  {
    ad_t *n;

    if(t == NULL)
      return;

    w_ad(t->l);

    n = TREAP_ENTRY(t);

    in.s_addr = htonl(n->addr);
    title_font_align(
      image, 14, 0, h_pos + 4, colour_white, inet_ntoa(in), HIST_X - 5);

    a2 = a1 = BAR_START;
    draw_line(n->voldr.I, colour_red);
    draw_line(n->voldr.O, colour_blue);
    draw_line(n->voldr.u, colour_purple);
    draw_line(n->voldr.i, colour_cyan);
    draw_line(n->voldr.o, colour_green);
    h_pos += BAR_SPACE;

    w_ad(t->r);
  }

 /*
  * start
  */
  printf("constructing graph- %s\n", filename);
  no_entries = 0;
  max_vol = 0;
  find_max(bt);
  graph_init();

  sprintf(str, "iptrk R%s- Address domain histogram", REVISION);
  title_font(image, 15, 5, 20, colour_white, str);
  title_font(image, 15, 5, 40, colour_white, ctime(&t));

  gdImageSetThickness(image, 10);
  h_pos = HIST_Y + 8;
  w_ad(bt);

  sprintf(str, IPTRK_GRAPHS_PATH "/%s", filename);
  fp = fopen(str, "w");
  if(fp != NULL)
  {
    gdImagePng(image, fp);
    fclose(fp);
  }
  else
  {
    log_e(__func__, "::fopen(png)");
    exit(-1);
  }

  gdImageDestroy(image);

 /*
  * add html listing
  */
  strcpy(html_entry.fn, filename);
  html_entry.t = t;
  html_entry.v = max_vol2;
  ll_append_unsorted(&html_ll, &html_entry, sizeof(html_t));
}

/**************************************************************************
*
**************************************************************************/
PRIVATE void construct_graphs(void)
{
  char          str[STR_LEN + 1], str2[STR_LEN + 1];
  double        high_vol_list = 0;
  unsigned long hash;
  FILE          *fp2;

  void w_td(treap_t *t)
  {
    td_t *n;

    if(t == NULL)
      return;

    w_td(t->l);

    n = TREAP_ENTRY(t);
    sprintf(str, "%lx.%lx.png", n->time, hash);
    make_a_graph(n->ad_bt, n->time, str);

    sprintf(str2, IPTRK_GRAPHS_PATH "/%s.html", str);
    fp2 = fopen(str2, "w");
    if(fp2 == NULL)
    {
      log_e(__func__, "::fopen(.png.html)");
      exit(-1);
    }

    fprintf(
      fp2,
      "<html>"
      "<title>iptrk graph stats (Address Domain)</title>"
      "<body TEXT=\"#ffffff\" BGCOLOR=\"#000000\" LINK=\"#8888FF\" VLINK=\"#FF0000\">"
      "<font color=\"#d0d0f0\">"
      "<center><img src=\"/iptrk.png\"><p>"
      "<font color=\"#80f0a0\">"
      "iptrk graph stats (Time Domain) for %s"
      "<hr><p>"
      "<img src=\"" IPTRK_GRAPHS_ABS "/%s\">", str, str);

    fprintf(
      fp2,
      "<hr>"
      "</center>"
      "</body>"
      "</html>");

    fclose(fp2);

    w_td(t->r);
  }

  void find_high_vol(ll_t *n)
  {
    html_t *e;

    if(n == NULL)
      return;

    e = LL_ENTRY(n);
    if(e->v > high_vol_list)
      high_vol_list = e->v;

    find_high_vol(n->next);
  }

  void print_html(ll_t *n)
  {
    html_t *e;

    if(n == NULL)
      return;

    e = LL_ENTRY(n);
    fprintf(
      web_fp,
      "<tr><th align=left><a href=\"" IPTRK_GRAPHS_ABS
      "/%s.html\">%s - (scale: %.2f Mb)</a></th>"
      "<th align=left><img src=\"/bar.png\" height=15 width=%u></th></tr>",
      e->fn, ctime(&(e->t)), e->v,
      (unsigned int)((e->v * BAR_LENGTH) / high_vol_list));

    print_html(n->next);
  }

 /*
  * start
  */
  hash = random() * timer * getpid();
  web_fp = fopen(IPTRK_GRAPHS_PATH "/index.html", "w");
  if(web_fp == NULL)
  {
    log_e(__func__, "::fopen(index.html)");
    exit(-1);
  }

  fprintf(
    web_fp,
    "<html>"
    "<title>Generated graphs (Address Domain)</title>"
    "<body><center>"
    "<img src=\"/iptrk.png\"><p>"
    "<font color=\"#80f0a0\">"
    "Generated graphs (Address Domain)"
    "<hr><table border=0>");

  w_td(td_bt);
  find_high_vol(html_ll);
  print_html(html_ll);
  fprintf(web_fp, "</table><hr></center></body></html>\n");
  fclose(web_fp);
}

/**************************************************************************
*
**************************************************************************/
int main(int argc, char *argv[])
{
  int  x;
  char str[STR_LEN + 1];

  welcome();
  log_m = cgi_log_m;
  timer = start_time = end_time = time(NULL);
  for(x = 0; x < argc; x++)
  {
    if(argv[x][0] == '-')
    {
      switch(argv[x][1])
      {
        case 'h':
          printf(
            "Usage: iptrkga <options>\n"
            " -h                    Display this help info\n"
            " -f <ip_log directory> ip_log's to interrogate\n"
            " -t <time>             Start time (Default = Epoch)\n"
            " -n <time>             End time   (Default = Current time)\n"
            "    (Nb/ Time format can be either YYYY:MM:DD:HH:MM or HH:MM)\n"
            "\n"
          );
          return 0;

        case 'f':
          strncpy2(ip_log_path, parse_cmdline(argc, argv, x), STR_LEN);
          break;

        case 't':
          start_time = datestr2time(parse_cmdline(argc, argv, x));
          break;

        case 'n':
          end_time = datestr2time(parse_cmdline(argc, argv, x));
          break;
      }
    }
  }

  printf(
    "\n"
    "ip_log path: %s\n", ip_log_path);
  format_date_entry(start_time, str);
  printf(
    " start time: %s\n", str);
  format_date_entry(end_time, str);
  printf(
    "   end time: %s\n", str);

  delete_old_pngs();
  parse_logs();
  construct_graphs();
  printf("done !\n");

  return 0;
}
